;;; -*- Mode: Lisp; Package: EDITOR -*-
;;; multiple-replace.l

;; Copyright (C) 2001-2005 OHKUBO Hiroshi <konata_o@hotmail.com>

;; Author: OHKUBO Hiroshi <konata_o@hotmail.com>
;; Time-stamp: <2005/04/13 00:15:00 +0900>

;;; Commentary:

;; Description:
;;
;;  ̒us܂B
;;

;; Installation:
;;
;;      1. A[JCuWJ multiple-replace.l  $XYZZY/site-lisp 
;;         Rs[܂B
;;
;;      2. KvȂ΃oCgRpC܂B
;;
;;              M-x byte-compile-file
;;              Byte compile file: $XYZZY/site-lisp/multiple-replace.l
;;
;;      3. ~/.xyzzy ܂ $XYZZY/site-lisp/siteinit.l Ɉȉ̃R[h
;;         ǉ܂B
;;
;;              (require "multiple-replace")
;;
;;      4. L̐ݒ𔽉f邽߂ɁAxyzzy ċN܂B
;;         siteinit.l ɋLqꍇ Ctrl L[ Shift L[Ȃ
;;         xyzzy ċNA_vt@Cč\z܂B
;;

;; Uninstallation:
;;
;;      1. ESC ESC (ed::multiple-replace-uninstall) ƃ^CvA
;;         multiple-replace.l ֘Ȁ xyzzy 폜܂B
;;
;;      2. multiple-replace.l ɊւLq폜܂B
;;
;;      3. siteinit.l ɋLqĂꍇ Ctrl L[ Shift L[
;;         Ȃ xyzzy ċNA_vt@Cč\z܂B
;;

;; Changes:
;;
;;      Wed, 13 Apr 2005 00:14:57 +0900
;;          Emultiple-replace-dialog ́uݒv͂ĂȂ
;;            ԂŁuۑvꍇA̎_̓ݒ薼肷
;;            悤ɊgB
;;
;;      Mon, 11 Apr 2005 21:22:43 +0900
;;          Exyzzy 0.2.2.234 Ή
;;            - "lr oSVbN"  "MS UI Gothic"
;;            - pJiSpJiɕύXB
;;            - multiple-replace-insert-menu-items XVB
;;              uuvj[ǉʒũftHg
;;              uuv̉ɁB
;;          ERgCB
;;
;;      Sun, 21 Dec 2003 08:20:37 +0900
;;          EsڂɌ͂ȂꍇAύX
;;            łȂ_CB
;;          Emenu ֘ACB
;;          Euninstall ֘AǉB
;;
;;      Sat, 04 Oct 2003 03:15:19 +0900
;;          EǗǉB
;;            ƂȂݒ̕ϐ multiple-replace Ǝ̂̂p悤ɁB
;;
;;      Mon, 14 Apr 2003 02:38:54 +0900
;;          Eō쐬B
;;

;; Exemption clause:
;;
;;      肪NĂ҂͈؊֒mӔC𕉂܂B
;;      g̐ӔCɂĂgB
;;

;;; Code:

(provide "multiple-replace")

(in-package "editor")

(require "search")
(require "dialogs")

(export '(multiple-replace-string
          query-multiple-replace
          multiple-replace-regexp
          query-multiple-replace-regexp
          multiple-replace-dialog
          multiple-replace-dialog-set-app-menu))

(define-history-variable *last-multiple-search-replace-string-list* nil)
(define-history-variable *last-multiple-search-replace-regexp-list* nil)
; (((pattern1 . replace1) (pattern2 . replace2) ...) ((decode-pattern1 . decode-replace1) ...))
(define-history-variable *last-multiple-search-replace-string-list-pair* nil)

(define-history-variable *last-multiple-replace-setting* nil)
; ((("hisitory-name" . string) ("setting" . sexp)) ...)
(define-history-variable *multiple-replace-setting-history* nil)

(defconstant *multiple-replace-dialog-number-of-input-min* 1)
(defvar *multiple-replace-dialog-number-of-input* 5)
(defvar *multiple-replace-pattern-replacement-list* nil)
(defvar *multiple-replace-case-fold-search* *case-fold-search*)
(defvar *multiple-replace-word-search* *word-search*)
(defvar *multiple-replace-regexp-search* *regexp-search*)
(defvar *multiple-replace-understand-escape-sequences* *understand-escape-sequences*)
(defvar *multiple-replace-replace-whole-buffer* *replace-whole-buffer*)

(defun perform-multiple-replace (pattern-replacement-list query regexp interactive-p noerror)
  (let ((literal (null regexp))
        (count 0)
        (nodup nil)
        (opoint (point))
        (not-found t)
        (def nil)
        (last-match nil)
        (last-match-char nil)
        (undo-bound (or interactive-p query))
        (pattern nil)
        (replacement nil)
        (point-pattern-list nil)
        (point-search nil))
    (undo-boundary)
    (when interactive-p
      (if regexp
          (setq *last-multiple-search-replace-regexp-list*
                pattern-replacement-list)
        (setq *last-multiple-search-replace-string-list*
              pattern-replacement-list))
      (let ((*multiple-replace-regexp-search* regexp))
        (declare (special *multiple-replace-regexp-search*))
        (setq *last-multiple-replace-setting* (multiple-replace-make-setting))))
    (if regexp
        (setq pattern-replacement-list
              (mapcar #'(lambda (pattern-replacement)
                          (cons
                           (if (stringp (car pattern-replacement))
                               (compile-regexp (car pattern-replacement) *multiple-replace-case-fold-search*)
                             (car pattern-replacement))
                           (cdr pattern-replacement)))
                      pattern-replacement-list)))
    (unless query
      (begin-wait-cursor))
    (loop
      (setq point-pattern-list nil)
      (save-excursion
        (setq search-point (point))
        (dolist (pattern (mapcar 'car pattern-replacement-list))
          (when (and (or (regexpp pattern) (stringp pattern))
                     (scan-buffer pattern
                                  :regexp regexp
                                  :no-dup nodup
                                  :last-match (cons last-match last-match-char)
                                  :case-fold *multiple-replace-case-fold-search*
                                  :left-bound *multiple-replace-word-search*
                                  :right-bound *multiple-replace-word-search*))
            (push (cons (point) pattern) point-pattern-list))
          (goto-char search-point)))
      (unless point-pattern-list
        (return))
      (setq pattern (cdar (stable-sort (nreverse point-pattern-list) '< :key 'car)))
      (setq replacement (cdr (assoc pattern pattern-replacement-list :test 'equal)))
      (scan-buffer pattern
                   :regexp regexp
                   :no-dup nodup
                   :last-match (cons last-match last-match-char)
                   :case-fold *multiple-replace-case-fold-search*
                   :left-bound *multiple-replace-word-search*
                   :right-bound *multiple-replace-word-search*)
      (if (and (eql (match-beginning 0) (match-end 0))
               (eql last-match (match-beginning 0)))
          (setq nodup t)
        (progn
          (setq not-found nil)
          (show-match)
          (unless (pos-visible-in-window-p (point))
            (recenter))
          (if query
              (progn
                (refresh-screen)
                (message (if (eq def 'help)
                             "(y)u, (n)XLbv, (!)cS, (u)AhD, (C-g)~, (.)~Ė߂"
                           (format nil "u?:  ~S" replacement)))
                (setq def (lookup-keymap *query-replace-map* (read-char *keyboard*) t)))
            (setq def 'automatic))
          (when (cond ((eq def 'act)
                       t)
                      ((eq def 'skip)
                       (setq nodup t)
                       nil)
                      ((eq def 'automatic)
                       (setq query nil)
                       t)
                      ((eq def 'undo)
                       (and (plusp count)
                            (undo)
                            (setq count (1- count)))
                       (setq nodup nil)
                       nil)
                      ((eq def 'quit-and-back)
                       (goto-char opoint)
                       (setq def 'quit)
                       (return))
                      ((eq def 'quit)
                       (return))
                      ((eq def 'recenter)
                       (recenter)
                       (setq nodup nil)
                       nil)
                      ((and (consp def)
                            (eq (car def) 'throw))
                       (throw (cdr def) count))
                      (t (setq def 'help)
                         (setq nodup nil)
                         nil))
            (setq last-match-char (unless (= (match-end 0) (point-min))
                                    (char-before (match-end 0))))
            (unless (replace-match replacement :literal literal)
              (return))
            (setq last-match (point))
            (setq nodup (= (match-beginning 0) (match-end 0)))
            (when (and undo-bound query)
              (undo-boundary))
            (setq count (1+ count)))))
      (when (eobp)
        (return)))
    (unless query
      (end-wait-cursor))
    (hide-match)
    (unless noerror
      (and (eq def 'quit)
           (quit))
      (and not-found
           (error "񂪌܂")))
    (when (and interactive-p
               (null *executing-macro*))
      (message "~du܂" count))
    count))

(defun multiple-replace-string (pattern-replacement-list &optional noerror)
  (interactive
      (let (pattern replacement pat-rep-list (i 1))
        (loop
          (setq pattern
                (read-string (format nil "Replace string (~D): " i)
                             :default (elt (mapcar 'car *last-multiple-search-replace-string-list*) (1- i))
                             :history 'search))
          (if (string= pattern "")
              (return)
            (progn
              (setq replacement
                    (read-string (format nil "with (~D): " i)
                                 :default (elt (mapcar 'cdr *last-multiple-search-replace-string-list*) (1- i))
                                 :history 'search))
              (push (cons pattern replacement) pat-rep-list)))
          (incf i))
        (list (nreverse pat-rep-list))))
  (when buffer-read-only
    (error "݋֎~obt@ł"))
  (perform-multiple-replace pattern-replacement-list nil nil (interactive-p) noerror))

(defun query-multiple-replace (pattern-replacement-list &optional noerror)
  (interactive
      (let (pattern replacement pat-rep-list (i 1))
        (loop
          (setq pattern
                (read-string (format nil "Query replace (~D): " i)
                             :default (elt (mapcar 'car *last-multiple-search-replace-string-list*) (1- i))
                             :history 'search))
          (if (string= pattern "")
              (return)
            (progn
              (setq replacement
                    (read-string (format nil "with (~D): " i)
                                 :default (elt (mapcar 'cdr *last-multiple-search-replace-string-list*) (1- i))
                                 :history 'search))
              (push (cons pattern replacement) pat-rep-list)))
          (incf i))
        (list (nreverse pat-rep-list))))
  (when buffer-read-only
    (error "݋֎~obt@ł"))
  (perform-multiple-replace pattern-replacement-list t nil (interactive-p) noerror))

(defun multiple-replace-regexp (pattern-replacement-list &optional noerror)
  (interactive
      (let (pattern replacement pat-rep-list (i 1))
        (loop
          (setq pattern
                (read-string (format nil "Replace regexp (~D): " i)
                             :default (elt (mapcar 'car *last-multiple-search-replace-regexp-list*) (1- i))
                             :history 'search))
          (if (string= pattern "")
              (return)
            (progn
              (setq replacement
                    (read-string (format nil "with (~D): " i)
                                 :default (elt (mapcar 'cdr *last-multiple-search-replace-regexp-list*) (1- i))
                                 :history 'search))
              (push (cons pattern replacement) pat-rep-list)))
          (incf i))
        (list (nreverse pat-rep-list))))
  (when buffer-read-only
    (error "݋֎~obt@ł"))
  (perform-multiple-replace pattern-replacement-list nil t (interactive-p) noerror))

(defun query-multiple-replace-regexp (pattern-replacement-list &optional noerror)
  (interactive
      (let (pattern replacement pat-rep-list (i 1))
        (loop
          (setq pattern
                (read-string (format nil "Query replace regexp (~D): " i)
                             :default (elt (mapcar 'car *last-multiple-search-replace-regexp-list*) (1- i))
                             :history 'search))
          (if (string= pattern "")
              (return)
            (progn
              (setq replacement
                    (read-string (format nil "with (~D): " i)
                                 :default (elt (mapcar 'cdr *last-multiple-search-replace-regexp-list*) (1- i))
                                 :history 'search))
              (push (cons pattern replacement) pat-rep-list)))
          (incf i))
        (list (nreverse pat-rep-list))))
  (when buffer-read-only
    (error "݋֎~obt@ł"))
  (perform-multiple-replace pattern-replacement-list t t (interactive-p) noerror))


(defun multiple-replace-get-setting-item (setting name &optional default)
  (let (lst)
    (if (setq lst (assoc name setting :test 'string=))
        (cadr lst)
      default)))
(defun multiple-replace-load-setting (setting)
  (setq *multiple-replace-case-fold-search* (multiple-replace-get-setting-item setting "case-fold-search"))
  (setq *multiple-replace-word-search* (multiple-replace-get-setting-item setting "word-search"))
  (setq *multiple-replace-regexp-search* (multiple-replace-get-setting-item setting "regexp-search"))
  (setq *multiple-replace-understand-escape-sequences*
        (multiple-replace-get-setting-item setting "understand-escape-sequences"))
  (setq *multiple-replace-replace-whole-buffer*
        (multiple-replace-get-setting-item setting "replace-whole-buffer"))
  (setq *multiple-replace-dialog-number-of-input*
        (multiple-replace-get-setting-item setting "dialog-number-of-input"
                                           *multiple-replace-dialog-number-of-input*))
  (if *multiple-replace-regexp-search*
      (setq *last-multiple-search-replace-regexp-list*
            (multiple-replace-get-setting-item setting "pattern-replacement-list"))
    (setq *last-multiple-search-replace-string-list*
          (multiple-replace-get-setting-item setting "pattern-replacement-list")))
  (setq *last-multiple-search-replace-string-list-pair*
        (multiple-replace-get-setting-item setting "search-replace-string-list-pair"))
  nil)
(defun multiple-replace-make-setting ()
  (list (list (copy-string "pattern-replacement-list")
              (if *multiple-replace-regexp-search*
                  *last-multiple-search-replace-regexp-list*
                *last-multiple-search-replace-string-list*))
        (list (copy-string "case-fold-search") *multiple-replace-case-fold-search*)
        (list (copy-string "word-search") *multiple-replace-word-search*)
        (list (copy-string "regexp-search") *multiple-replace-regexp-search*)
        (list (copy-string "understand-escape-sequences") *multiple-replace-understand-escape-sequences*)
        (list (copy-string "replace-whole-buffer") *multiple-replace-replace-whole-buffer*)
        (list (copy-string "dialog-number-of-input") *multiple-replace-dialog-number-of-input*)
        (list (copy-string "search-replace-string-list-pair") *last-multiple-search-replace-string-list-pair*)))

(defun multiple-replace-dialog (&key (setting nil sv-setting) history-name)
  "𓯎ɒu"
  (interactive)
  (let (dialog-template
        number-of-input
        controls controls-initializers controls-handlers
        search-symbol-list replace-symbol-list
        history
        (history-name-list (mapcar
                            #'(lambda (h)
                                (multiple-replace-get-setting-item h "history-name"))
                            *multiple-replace-setting-history*)))
    (let (pos)
      (if (and (stringp history-name)
               (setq pos (position history-name *multiple-replace-setting-history*
                                   :key #'(lambda (h) (multiple-replace-get-setting-item h "history-name"))
                                   :test 'equal)))
          (setq setting (multiple-replace-get-setting-item
                         (nth pos *multiple-replace-setting-history*) "setting"))
        (progn
          (setq history-name nil)
          (if (and (null setting) (not sv-setting))
              (setq setting *last-multiple-replace-setting*)))))
    (multiple-replace-load-setting setting)
    (setq number-of-input (max *multiple-replace-dialog-number-of-input-min*
                               (if (and (numberp *multiple-replace-dialog-number-of-input*)
                                        (integerp *multiple-replace-dialog-number-of-input*)
                                        (plusp *multiple-replace-dialog-number-of-input*))
                                   *multiple-replace-dialog-number-of-input* 0)))

    ;; search, replace p symbol ̐
    (dotimes (i number-of-input)
      (push (make-symbol (concat "search" (format nil "~D" i))) search-symbol-list)
      (push (make-symbol (concat "replace" (format nil "~D" i))) replace-symbol-list))
    (setq search-symbol-list (nreverse search-symbol-list))
    (setq replace-symbol-list (nreverse replace-symbol-list))
    
    (push `(:button nil "" #x50020007
            2 2 310 ,(+ 22 (* 14 number-of-input))) controls)
    (push '(:static nil ""      #x50020001   16 10 140 8) controls)
    (push '(:static nil "u"      #x50020001  168 10 140 8) controls)
    (dotimes (i number-of-input)
      (if (< i 9)
          (push `(:static nil ,(format nil "&~D:" (1+ i)) #x50020002
                  7 ,(+ 22 (* 14 i)) 8 8) controls)
        (push `(:static nil ,(multiple-value-bind (dec num)
                                 (floor (1+ i) 10)
                               (format nil "~D&~D:" dec num)) #x50020002
                7 ,(+ 22 (* 14 i)) 8 8) controls))
      (push `(:static nil "" #x50020000
              159 ,(+ 22 (* 14 i)) 8 8) controls)
      (push `(:combobox ,(elt search-symbol-list i) nil  #x50210842
              16 ,(+ 19 (* 14 i)) 140 96) controls)
      (push `(:combobox ,(elt replace-symbol-list i) nil #x50210842
              168 ,(+ 19 (* 14 i)) 140 96) controls))
    (push `(:button case-fold "啶ʂ(&C)" #x50010006
            39 ,(+ 30 (* 14 number-of-input)) 113 10) controls)
    (push `(:button word "PPʂŌ(&W)" #x50010003
            39 ,(+ 43 (* 14 number-of-input)) 97 10) controls)
    (push `(:button regexp "K\(&E)" #x50010003
            39 ,(+ 56 (* 14 number-of-input)) 63 10) controls)
    (push `(:button escseq "GXP[vV[PX𗝉(&Y)" #x50010003
            39 ,(+ 69 (* 14 number-of-input)) 113 10) controls)
    (push `(:button whole "obt@̐擪(&O)" #x50010003
            39 ,(+ 82 (* 14 number-of-input)) 89 10) controls)

    (push `(:button query "mF(&Q)" #x50010001
            185 ,(+ 30 (* 14 number-of-input)) 50 14) controls)
    (push `(:button all "SĒu(&A)" #x50010000
            185 ,(+ 47 (* 14 number-of-input)) 50 14) controls)
    (push `(:button apply "͐ύX(&I)" #x50010000
            185 ,(+ 64 (* 14 number-of-input)) 50 14) controls)
    (push `(:static nil "(&N)" #x50020000
            264 ,(+ 67 (* 14 number-of-input)) 10 8) controls)
    (push `(:edit resize-num nil #x50812086
            238 ,(+ 64 (* 14 number-of-input)) 26 14) controls)
    (push `(:spin resize-numspin nil #x500000b6
            254 ,(+ 64 (* 14 number-of-input)) 10 14) controls)
    (push `(:button IDCANCEL "LZ" #x50010000
            185 ,(+ 81 (* 14 number-of-input)) 50 14) controls)

    (push `(:static nil "ݒ(&T):" #x50020001
            120 ,(+ 102 (* 14 number-of-input)) 30 8) controls)
    (push `(:combobox history nil #x50210842
            150 ,(+ 100 (* 14 number-of-input)) 80 80) controls)
    (push `(:button history-save "ۑ"  #x50010000
            234 ,(+ 100 (* 14 number-of-input)) 24 14) controls)
    (push `(:button history-load "Ǎ"  #x50010000
            260 ,(+ 100 (* 14 number-of-input)) 24 14) controls)
    (push `(:button history-del "폜"  #x50010000
            286 ,(+ 100 (* 14 number-of-input)) 24 14) controls)

    (setq controls (nreverse controls))

    ;; controls-initializers ̐
    (push `(case-fold . ,(cfs2dialog *multiple-replace-case-fold-search*)) controls-initializers)
    (push `(word . ,*multiple-replace-word-search*) controls-initializers)
    (push `(regexp . ,*multiple-replace-regexp-search*) controls-initializers)
    (push `(escseq . ,*multiple-replace-understand-escape-sequences*) controls-initializers)
    (push `(whole . ,*multiple-replace-replace-whole-buffer*) controls-initializers)
    (let (pattern-replacement-list)
      (if *multiple-replace-regexp-search*
          (setq pattern-replacement-list
                *last-multiple-search-replace-regexp-list*)
        (setq pattern-replacement-list
              *last-multiple-search-replace-string-list*))

      ; decode-escape-sequence ̕Ɠl̏ꍇ decode O̕
      (if (equalp (cdr *last-multiple-search-replace-string-list-pair*)
                  pattern-replacement-list)
          (setq pattern-replacement-list (car *last-multiple-search-replace-string-list-pair*)))
      (dotimes (i (min number-of-input (length pattern-replacement-list)))
        (push `(,(elt search-symbol-list i) . ,(car (elt pattern-replacement-list i)))
              controls-initializers)
        (push `(,(elt replace-symbol-list i) . ,(cdr (elt pattern-replacement-list i)))
              controls-initializers))
      (dotimes (i number-of-input)
        (push `(,(elt search-symbol-list i) . ,*minibuffer-search-string-history*)
              controls-initializers)
        (push `(,(elt replace-symbol-list i) . ,*minibuffer-search-string-history*)
              controls-initializers)))
    (push `(resize-num . ,(format nil "~D" *multiple-replace-dialog-number-of-input*))
          controls-initializers)

    (push (cons 'history history-name) controls-initializers)
    (push (cons 'history history-name-list) controls-initializers)

    (setq controls-initializers (nreverse controls-initializers))

    ;; controls-handlers ̐
    (push '(word :disable (regexp)) controls-handlers)
    (push `(resize-num :min ,*multiple-replace-dialog-number-of-input-min*
                       :max 99 :range-error
                       ,(format nil "~D ȏ 99 ȉw肵Ă"
                                *multiple-replace-dialog-number-of-input-min*)
                       :type integer :type-error "l͂Ă")
          controls-handlers)
    (push `(resize-numspin :min ,*multiple-replace-dialog-number-of-input-min*
                           :max 99) controls-handlers)
;    (push `(,(elt search-symbol-list 0) :non-null "͂"
;            :enable (query all)) controls-handlers)

    (setq controls-handlers (nreverse controls-handlers))

    ;;
    (setq dialog-template
          `(dialog 0 0 316 ,(+ 119 (* 14 number-of-input))
                   (:caption ,(if (stringp history-name)
                                  (concat "̒u - [" history-name "]")
                                "̒u"))
                   (:font 9 "MS UI Gothic")
                   (:control
                    ,@controls)))
    (multiple-value-bind (result data)
        (dialog-box dialog-template
                    controls-initializers
                    controls-handlers)
      (when result
        (let (pattern-replacement-list j pattern replace)
          (dotimes (i number-of-input)
            (setq j (- number-of-input 1 i))
            (setq pattern (cdr (assoc (elt search-symbol-list j) data)))
            (setq replace (cdr (assoc (elt replace-symbol-list j) data)))
            (push (cons pattern replace) pattern-replacement-list)
            (if (stringp pattern)
                (add-history pattern '*minibuffer-search-string-history*))
            (if (stringp replace)
                (add-history replace '*minibuffer-search-string-history*)))

          (setq *multiple-replace-case-fold-search* (dialog2cfs (cdr (assoc 'case-fold data))))
          (setq *multiple-replace-word-search* (cdr (assoc 'word data)))
          (setq *multiple-replace-regexp-search* (cdr (assoc 'regexp data)))
          (setq *multiple-replace-understand-escape-sequences* (cdr (assoc 'escseq data)))
          (setq *multiple-replace-replace-whole-buffer* (cdr (assoc 'whole data)))
          (setq *multiple-replace-dialog-number-of-input* (cdr (assoc 'resize-num data)))
          (if *multiple-replace-understand-escape-sequences*
              (setq *last-multiple-search-replace-string-list-pair*
                    (cons (copy-tree pattern-replacement-list)
                          (setq pattern-replacement-list
                                (mapcar #'(lambda (pattern-replacement)
                                            (cons (decode-escape-sequence
                                                   (car pattern-replacement)
                                                   *multiple-replace-regexp-search*)
                                                  (decode-escape-sequence
                                                   (cdr pattern-replacement)
                                                   *multiple-replace-regexp-search*)))
                                        pattern-replacement-list))))
            (setq *last-multiple-search-replace-string-list-pair* nil))
          ;; Ǘ̓sAperform-multiple-replace łȂAłۑ
          (if *multiple-replace-regexp-search*
              (setq *last-multiple-search-replace-regexp-list*
                    pattern-replacement-list)
            (setq *last-multiple-search-replace-string-list*
                  pattern-replacement-list))

          (setq *last-multiple-replace-setting* (multiple-replace-make-setting))
          (setq history-name (cdr (assoc 'history data)))

          (cond
           ;; Kp
           ((eq result 'apply)
            (multiple-replace-dialog))
           ;; s
           ((or (eq result 'query) (eq result 'all))
            (when (cdr (assoc (elt search-symbol-list 0) data))
              (if *multiple-replace-replace-whole-buffer*
                  (goto-char (point-min)))
              (perform-multiple-replace pattern-replacement-list (eq result 'query)
                                        (and (null *multiple-replace-word-search*)
                                             *multiple-replace-regexp-search*)
                                        t nil)))
           ;; 폜
           ((eq result 'history-del)
            (let (pos)
              (if (setq pos
                        (position history-name *multiple-replace-setting-history*
                                  :key #'(lambda (h) (multiple-replace-get-setting-item h "history-name"))
                                  :test 'equal))
                  (setq *multiple-replace-setting-history*
                        (append (subseq *multiple-replace-setting-history* 0 pos)
                                (subseq *multiple-replace-setting-history* (1+ pos))))))
            (multiple-replace-dialog))
           ;; ۑ
           ((eq result 'history-save)
            (let (pos)
              (when (string-match "^[ \t]*$" history-name)
                (setq history-name (format-date-string "%Y/%m/%d %H:%M:%S")))
              (if (setq pos
                        (position history-name *multiple-replace-setting-history*
                                  :key #'(lambda (h) (multiple-replace-get-setting-item h "history-name"))
                                  :test 'equal))
                  (setq *multiple-replace-setting-history*
                        (append (subseq *multiple-replace-setting-history* 0 pos)
                                (subseq *multiple-replace-setting-history* (1+ pos)))))
              (push (list (list (copy-string "history-name") history-name)
                          (list (copy-string "setting") *last-multiple-replace-setting*))
                    *multiple-replace-setting-history*))
            (multiple-replace-dialog :history-name history-name))
           ;; Ǎ
           ((eq result 'history-load)
            (let (pos history)
              (if (setq pos
                        (position history-name *multiple-replace-setting-history*
                                  :key #'(lambda (h) (multiple-replace-get-setting-item h "history-name"))
                                  :test 'equal))
                  (progn
                    (setq history (nth pos *multiple-replace-setting-history*))
                    (setq *last-multiple-replace-setting* (multiple-replace-get-setting-item history "setting"))
                    (setq *multiple-replace-setting-history*
                          (append (subseq *multiple-replace-setting-history* 0 pos)
                                  (subseq *multiple-replace-setting-history* (1+ pos))))
                    (push history *multiple-replace-setting-history*)
                    (multiple-replace-dialog :history-name history-name))
                (multiple-replace-dialog :setting nil)))))
          t)))))

(defun multiple-replace-insert-menu-items (&key menu pre-tag position head-sep tail-sep)
  (if (and (not (menup menu)) (menup *app-menu*))
      (setq menu (get-menu *app-menu* 'ed::search)))
  (when (menup menu)
    (multiple-replace-delete-menu menu)
    (unless (or pre-tag position)
      (setq pre-tag 'ed::replace-dialog))
    (when (and pre-tag
               (setq position (get-menu-position menu pre-tag)))
      (incf position))
    (unless (and (numberp position) (integerp position) (not (minusp position))
                 (get-menu menu position t))
      (setq position -1)
      (while (get-menu menu (incf position) t)))
    (decf position)
    (if (and head-sep
             (not (minusp position))
             (get-menu menu position t))
        (insert-menu-separator menu (incf position) 'multiple-replace-dialog-sep))
    (insert-menu-item menu (incf position) 'multiple-replace-dialog
                      "u(&Q)..." 'multiple-replace-dialog)
    (if (and tail-sep
             (get-menu menu (incf position) t))
        (insert-menu-separator menu position 'multiple-replace-dialog-sep))))

(defun multiple-replace-delete-menu (&optional menu)
  (if (and (not (menup menu)) (menup *app-menu*))
      (setq menu (get-menu *app-menu* 'ed::search)))
  (when (menup menu)
    (while (delete-menu menu 'multiple-replace-dialog))
    (while (delete-menu menu 'multiple-replace-dialog-sep))))

(defun multiple-replace-uninstall ()
  (let ((history-variables
         '(*last-multiple-search-replace-string-list*
           *last-multiple-search-replace-regexp-list*
           *last-multiple-search-replace-string-list-pair*
           *last-multiple-replace-setting*
           *multiple-replace-setting-history*)))
    (dolist (variable history-variables)
      (unregister-history-variable variable)))
  (multiple-replace-delete-menu))

(if (menup ed::*app-menu*)
    (multiple-replace-insert-menu-items)
  (add-hook 'ed::*init-app-menus-hook*
            #'(lambda () (multiple-replace-insert-menu-items))))

;;; multiple-replace.l ends here
